@@ -22,7 +22,7 @@
 
 /*----------------------------------------------------------------------------*/
 
-void flush_proc(struct work_struct *);
+int flush_daemon_proc(void *);
 void wait_for_flushing(struct wb_device *, u64 id);
 
 /*----------------------------------------------------------------------------*/
@@ -32,8 +32,10 @@ void flush_barrier_ios(struct work_struct *);
 
 /*----------------------------------------------------------------------------*/
 
+void update_nr_empty_segs(struct wb_device *);
 int writeback_daemon_proc(void *);
 void wait_for_writeback(struct wb_device *, u64 id);
+void mark_clean_seg(struct wb_device *, struct segment_header *seg);
 
 /*----------------------------------------------------------------------------*/
 
