@@ -249,7 +249,7 @@ static void free_ht(struct wb_device *wb)
 struct ht_head *ht_get_head(struct wb_device *wb, struct lookup_key *key)
 {
 	u32 idx;
-	div_u64_rem(key->sector, wb->htsize, &idx);
+	div_u64_rem(key->sector >> 3, wb->htsize, &idx);
 	return large_array_at(wb->htable, idx);
 }
 
@@ -277,6 +277,7 @@ void ht_register(struct wb_device *wb, struct ht_head *head,
 	hlist_del(&mb->ht_list);
 	hlist_add_head(&mb->ht_list, &head->ht_list);
 
+	BUG_ON(key->sector & 7); // should be 4KB aligned
 	mb->sector = key->sector;
 };
 
@@ -310,7 +311,7 @@ void discard_caches_inseg(struct wb_device *wb, struct segment_header *seg)
 static int read_superblock_header(struct superblock_header_device *sup,
 				  struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 	struct dm_io_request io_req_sup;
 	struct dm_io_region region_sup;
 
@@ -320,8 +321,8 @@ static int read_superblock_header(struct superblock_header_device *sup,
 	check_buffer_alignment(buf);
 
 	io_req_sup = (struct dm_io_request) {
+		WB_IO_READ,
 		.client = wb->io_client,
-		.bi_rw = READ,
 		.notify.fn = NULL,
 		.mem.type = DM_IO_KMEM,
 		.mem.ptr.addr = buf,
@@ -331,15 +332,15 @@ static int read_superblock_header(struct superblock_header_device *sup,
 		.sector = 0,
 		.count = 1,
 	};
-	r = wb_io(&io_req_sup, 1, &region_sup, NULL, false);
-	if (r)
+	err = wb_io(&io_req_sup, 1, &region_sup, NULL, false);
+	if (err)
 		goto bad_io;
 
 	memcpy(sup, buf, sizeof(*sup));
 
 bad_io:
 	mempool_free(buf, wb->buf_1_pool);
-	return r;
+	return err;
 }
 
 /*
@@ -348,28 +349,28 @@ bad_io:
  */
 static int audit_cache_device(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 	struct superblock_header_device sup;
-	r = read_superblock_header(&sup, wb);
-	if (r) {
+	err = read_superblock_header(&sup, wb);
+	if (err) {
 		DMERR("read_superblock_header failed");
-		return r;
+		return err;
 	}
 
 	wb->do_format = false;
 	if (le32_to_cpu(sup.magic) != WB_MAGIC ||
-	    wb->write_around_mode) { // write-around mode should discard all caches
+	    wb->write_around_mode) { /* write-around mode should discard all caches */
 		wb->do_format = true;
 		DMERR("Superblock Header: Magic number invalid");
 		return 0;
 	}
 
-	return r;
+	return err;
 }
 
 static int format_superblock_header(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 
 	struct dm_io_request io_req_sup;
 	struct dm_io_region region_sup;
@@ -385,8 +386,8 @@ static int format_superblock_header(struct wb_device *wb)
 	memcpy(buf, &sup, sizeof(sup));
 
 	io_req_sup = (struct dm_io_request) {
+		WB_IO_WRITE_FUA,
 		.client = wb->io_client,
-		.bi_rw = WRITE_FUA,
 		.notify.fn = NULL,
 		.mem.type = DM_IO_KMEM,
 		.mem.ptr.addr = buf,
@@ -396,13 +397,13 @@ static int format_superblock_header(struct wb_device *wb)
 		.sector = 0,
 		.count = 1,
 	};
-	r = wb_io(&io_req_sup, 1, &region_sup, NULL, false);
-	if (r)
+	err = wb_io(&io_req_sup, 1, &region_sup, NULL, false);
+	if (err)
 		goto bad_io;
 
 bad_io:
 	mempool_free(buf, wb->buf_1_pool);
-	return r;
+	return err;
 }
 
 struct format_segmd_context {
@@ -436,13 +437,13 @@ static void zeroing_complete(int read_err, unsigned long write_err, void *contex
  */
 static int do_zeroing_region(struct wb_device *wb, struct dm_io_region *region)
 {
-	int r;
+	int err;
 	struct zeroing_context zc;
 	zc.error = 0;
 	init_completion(&zc.complete);
-	r = dm_kcopyd_zero(wb->copier, 1, region, 0, zeroing_complete, &zc);
-	if (r)
-		return r;
+	err = dm_kcopyd_zero(wb->copier, 1, region, 0, zeroing_complete, &zc);
+	if (err)
+		return err;
 	wait_for_completion(&zc.complete);
 	return zc.error;
 }
@@ -459,7 +460,7 @@ static int zeroing_full_superblock(struct wb_device *wb)
 
 static int format_all_segment_headers(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 	struct dm_dev *dev = wb->cache_dev;
 	u32 i;
 
@@ -474,13 +475,11 @@ static int format_all_segment_headers(struct wb_device *wb)
 	atomic64_set(&context.count, wb->nr_segments);
 	context.err = 0;
 
-	/*
-	 * Submit all the writes asynchronously.
-	 */
+	/* Submit all the writes asynchronously. */
 	for (i = 0; i < wb->nr_segments; i++) {
 		struct dm_io_request io_req_seg = {
+			WB_IO_WRITE,
 			.client = wb->io_client,
-			.bi_rw = WRITE,
 			.notify.fn = format_segmd_endio,
 			.notify.context = &context,
 			.mem.type = DM_IO_KMEM,
@@ -491,12 +490,12 @@ static int format_all_segment_headers(struct wb_device *wb)
 			.sector = calc_segment_header_start(wb, i),
 			.count = (1 << 3),
 		};
-		r = wb_io(&io_req_seg, 1, &region_seg, NULL, false);
-		if (r)
+		err = wb_io(&io_req_seg, 1, &region_seg, NULL, false);
+		if (err)
 			break;
 	}
 
-	if (r)
+	if (err)
 		goto bad;
 
 	/* Wait for all the writes complete. */
@@ -505,12 +504,15 @@ static int format_all_segment_headers(struct wb_device *wb)
 
 	if (context.err) {
 		DMERR("I/O failed");
-		r = -EIO;
+		err = -EIO;
+		goto bad;
 	}
 
+	err = blkdev_issue_flush(dev->bdev, GFP_KERNEL, NULL);
+
 bad:
 	mempool_free(buf, wb->buf_8_pool);
-	return r;
+	return err;
 }
 
 /*
@@ -518,27 +520,22 @@ bad:
  */
 static int format_cache_device(struct wb_device *wb)
 {
-	int r = 0;
-	struct dm_dev *dev = wb->cache_dev;
-
-	r = zeroing_full_superblock(wb);
-	if (r) {
+	int err = zeroing_full_superblock(wb);
+	if (err) {
 		DMERR("zeroing_full_superblock failed");
-		return r;
+		return err;
 	}
-	r = format_superblock_header(wb); /* First 512B */
-	if (r) {
-		DMERR("format_superblock_header failed");
-		return r;
-	}
-	r = format_all_segment_headers(wb);
-	if (r) {
+	err = format_all_segment_headers(wb);
+	if (err) {
 		DMERR("format_all_segment_headers failed");
-		return r;
+		return err;
 	}
-	r = blkdev_issue_flush(dev->bdev, GFP_KERNEL, NULL);
-
-	return r;
+	err = format_superblock_header(wb); /* First 512B */
+	if (err) {
+		DMERR("format_superblock_header failed");
+		return err;
+	}
+	return err;
 }
 
 /*
@@ -549,30 +546,30 @@ static int format_cache_device(struct wb_device *wb)
  */
 static int might_format_cache_device(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 
-	r = audit_cache_device(wb);
-	if (r) {
+	err = audit_cache_device(wb);
+	if (err) {
 		DMERR("audit_cache_device failed");
-		return r;
+		return err;
 	}
 
 	if (wb->do_format) {
-		r = format_cache_device(wb);
-		if (r) {
+		err = format_cache_device(wb);
+		if (err) {
 			DMERR("format_cache_device failed");
-			return r;
+			return err;
 		}
 	}
 
-	return r;
+	return err;
 }
 
 /*----------------------------------------------------------------------------*/
 
 static int init_rambuf_pool(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 	size_t i;
 
 	wb->rambuf_pool = kmalloc(sizeof(struct rambuffer) * NR_RAMBUF_POOL, GFP_KERNEL);
@@ -587,17 +584,17 @@ static int init_rambuf_pool(struct wb_device *wb)
 			for (j = 0; j < i; j++) {
 				vfree(wb->rambuf_pool[j].data);
 			}
-			r = -ENOMEM;
+			err = -ENOMEM;
 			goto bad_alloc_data;
 		}
 		wb->rambuf_pool[i].data = alloced;
 	}
 
-	return r;
+	return err;
 
 bad_alloc_data:
 	kfree(wb->rambuf_pool);
-	return r;
+	return err;
 }
 
 static void free_rambuf_pool(struct wb_device *wb)
@@ -608,6 +605,13 @@ static void free_rambuf_pool(struct wb_device *wb)
 	kfree(wb->rambuf_pool);
 }
 
+struct rambuffer *get_rambuffer_by_id(struct wb_device *wb, u64 id)
+{
+	u32 tmp32;
+	div_u64_rem(id - 1, NR_RAMBUF_POOL, &tmp32);
+	return wb->rambuf_pool + tmp32;
+}
+
 /*----------------------------------------------------------------------------*/
 
 /*
@@ -617,19 +621,19 @@ static void free_rambuf_pool(struct wb_device *wb)
  */
 static int init_devices(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 
-	r = might_format_cache_device(wb);
-	if (r)
-		return r;
+	err = might_format_cache_device(wb);
+	if (err)
+		return err;
 
-	r = init_rambuf_pool(wb);
-	if (r) {
+	err = init_rambuf_pool(wb);
+	if (err) {
 		DMERR("init_rambuf_pool failed");
-		return r;
+		return err;
 	}
 
-	return r;
+	return err;
 }
 
 static void free_devices(struct wb_device *wb)
@@ -642,7 +646,7 @@ static void free_devices(struct wb_device *wb)
 static int read_superblock_record(struct superblock_record_device *record,
 				  struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 	struct dm_io_request io_req;
 	struct dm_io_region region;
 
@@ -653,8 +657,8 @@ static int read_superblock_record(struct superblock_record_device *record,
 	check_buffer_alignment(buf);
 
 	io_req = (struct dm_io_request) {
+		WB_IO_READ,
 		.client = wb->io_client,
-		.bi_rw = READ,
 		.notify.fn = NULL,
 		.mem.type = DM_IO_KMEM,
 		.mem.ptr.addr = buf,
@@ -664,15 +668,15 @@ static int read_superblock_record(struct superblock_record_device *record,
 		.sector = (1 << 11) - 1,
 		.count = 1,
 	};
-	r = wb_io(&io_req, 1, &region, NULL, false);
-	if (r)
+	err = wb_io(&io_req, 1, &region, NULL, false);
+	if (err)
 		goto bad_io;
 
 	memcpy(record, buf, sizeof(*record));
 
 bad_io:
 	mempool_free(buf, wb->buf_1_pool);
-	return r;
+	return err;
 }
 
 /*
@@ -682,8 +686,8 @@ static int read_whole_segment(void *buf, struct wb_device *wb,
 			      struct segment_header *seg)
 {
 	struct dm_io_request io_req = {
+		WB_IO_READ,
 		.client = wb->io_client,
-		.bi_rw = READ,
 		.notify.fn = NULL,
 		.mem.type = DM_IO_VMA,
 		.mem.ptr.addr = buf,
@@ -713,7 +717,7 @@ void prepare_segment_header_device(void *rambuffer,
 	struct segment_header_device *dest = rambuffer;
 	u32 i;
 
-	BUG_ON((src->length) != (wb->cursor - src->start_idx));
+	ASSERT((src->length) == (wb->cursor - src->start_idx));
 
 	for (i = 0; i < src->length; i++) {
 		struct metablock *mb = src->mb_array + i;
@@ -774,8 +778,8 @@ static int apply_metablock_device(struct wb_device *wb, struct segment_header *s
 				continue;
 
 			io_req = (struct dm_io_request) {
+				WB_IO_WRITE,
 				.client = wb->io_client,
-				.bi_rw = WRITE,
 				.notify.fn = NULL,
 				.mem.type = DM_IO_KMEM,
 				.mem.ptr.addr = wio.data + (i << 9),
@@ -825,8 +829,8 @@ static int read_segment_header(void *buf, struct wb_device *wb,
 			       struct segment_header *seg)
 {
 	struct dm_io_request io_req = {
+		WB_IO_READ,
 		.client = wb->io_client,
-		.bi_rw = READ,
 		.notify.fn = NULL,
 		.mem.type = DM_IO_KMEM,
 		.mem.ptr.addr = buf,
@@ -845,7 +849,7 @@ static int read_segment_header(void *buf, struct wb_device *wb,
  */
 static int do_find_max_id(struct wb_device *wb, u64 *max_id)
 {
-	int r = 0;
+	int err = 0;
 	u32 k;
 
 	void *buf = mempool_alloc(wb->buf_8_pool, GFP_KERNEL);
@@ -857,8 +861,8 @@ static int do_find_max_id(struct wb_device *wb, u64 *max_id)
 	for (k = 0; k < wb->nr_segments; k++) {
 		struct segment_header *seg = segment_at(wb, k);
 		struct segment_header_device *header;
-		r = read_segment_header(buf, wb, seg);
-		if (r)
+		err = read_segment_header(buf, wb, seg);
+		if (err)
 			goto out;
 
 		header = buf;
@@ -867,7 +871,7 @@ static int do_find_max_id(struct wb_device *wb, u64 *max_id)
 	}
 out:
 	mempool_free(buf, wb->buf_8_pool);
-	return r;
+	return err;
 }
 
 static int find_max_id(struct wb_device *wb, u64 *max_id)
@@ -897,7 +901,7 @@ static int find_max_id(struct wb_device *wb, u64 *max_id)
  */
 static int do_apply_valid_segments(struct wb_device *wb, u64 *max_id)
 {
-	int r = 0;
+	int err = 0;
 	struct segment_header *seg;
 	struct segment_header_device *header;
 	u32 i, start_idx;
@@ -918,8 +922,8 @@ static int do_apply_valid_segments(struct wb_device *wb, u64 *max_id)
 		div_u64_rem(i, wb->nr_segments, &k);
 		seg = segment_at(wb, k);
 
-		r = read_whole_segment(rambuf, wb, seg);
-		if (r)
+		err = read_whole_segment(rambuf, wb, seg);
+		if (err)
 			break;
 
 		header = rambuf;
@@ -947,15 +951,15 @@ static int do_apply_valid_segments(struct wb_device *wb, u64 *max_id)
 		}
 
 		/* This segment is correct and we apply */
-		r = apply_segment_header_device(wb, seg, header);
-		if (r)
+		err = apply_segment_header_device(wb, seg, header);
+		if (err)
 			break;
 
 		*max_id = le64_to_cpu(header->id);
 	}
 
 	vfree(rambuf);
-	return r;
+	return err;
 }
 
 static int apply_valid_segments(struct wb_device *wb, u64 *max_id)
@@ -973,17 +977,18 @@ static int apply_valid_segments(struct wb_device *wb, u64 *max_id)
 
 static int infer_last_writeback_id(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 
+	u64 inferred_last_writeback_id;
 	u64 record_id;
-	struct superblock_record_device uninitialized_var(record);
-	r = read_superblock_record(&record, wb);
-	if (r)
-		return r;
 
-	atomic64_set(&wb->last_writeback_segment_id,
-		atomic64_read(&wb->last_flushed_segment_id) > wb->nr_segments ?
-		atomic64_read(&wb->last_flushed_segment_id) - wb->nr_segments : 0);
+	struct superblock_record_device uninitialized_var(record);
+	err = read_superblock_record(&record, wb);
+	if (err)
+		return err;
+
+	inferred_last_writeback_id =
+		SUB_ID(atomic64_read(&wb->last_flushed_segment_id), wb->nr_segments);
 
 	/*
 	 * If last_writeback_id is recorded on the super block
@@ -991,10 +996,15 @@ static int infer_last_writeback_id(struct wb_device *wb)
 	 * written back before.
 	 */
 	record_id = le64_to_cpu(record.last_writeback_segment_id);
-	if (record_id > atomic64_read(&wb->last_writeback_segment_id))
-		atomic64_set(&wb->last_writeback_segment_id, record_id);
+	if (record_id > inferred_last_writeback_id) {
+		u64 id;
+		for (id = inferred_last_writeback_id + 1; id <= record_id; id++)
+			mark_clean_seg(wb, get_segment_header_by_id(wb, id));
+		inferred_last_writeback_id = record_id;
+	}
 
-	return r;
+	atomic64_set(&wb->last_writeback_segment_id, inferred_last_writeback_id);
+	return err;
 }
 
 /*
@@ -1011,28 +1021,31 @@ static int infer_last_writeback_id(struct wb_device *wb)
  */
 static int replay_log_on_cache(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 
 	u64 max_id;
-	r = find_max_id(wb, &max_id);
-	if (r) {
+	err = find_max_id(wb, &max_id);
+	if (err) {
 		DMERR("find_max_id failed");
-		return r;
+		return err;
 	}
 
-	r = apply_valid_segments(wb, &max_id);
-	if (r) {
+	err = apply_valid_segments(wb, &max_id);
+	if (err) {
 		DMERR("apply_valid_segments failed");
-		return r;
+		return err;
 	}
 
 	/* Setup last_flushed_segment_id */
 	atomic64_set(&wb->last_flushed_segment_id, max_id);
 
+	/* Setup last_queued_segment_id */
+	atomic64_set(&wb->last_queued_segment_id, max_id);
+
 	/* Setup last_writeback_segment_id */
 	infer_last_writeback_id(wb);
 
-	return r;
+	return err;
 }
 
 /*
@@ -1050,12 +1063,12 @@ static void prepare_first_seg(struct wb_device *wb)
  */
 static int recover_cache(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 
-	r = replay_log_on_cache(wb);
-	if (r) {
+	err = replay_log_on_cache(wb);
+	if (err) {
 		DMERR("replay_log_on_cache failed");
-		return r;
+		return err;
 	}
 
 	prepare_first_seg(wb);
@@ -1123,7 +1136,7 @@ static void free_writeback_ios(struct wb_device *wb)
  */
 int try_alloc_writeback_ios(struct wb_device *wb, size_t nr_batch, gfp_t gfp)
 {
-	int r = 0;
+	int err = 0;
 	size_t i;
 
 	struct writeback_segment **writeback_segs = kzalloc(
@@ -1154,9 +1167,9 @@ int try_alloc_writeback_ios(struct wb_device *wb, size_t nr_batch, gfp_t gfp)
 
 	/* And then swap by new values */
 	wb->writeback_segs = writeback_segs;
-	wb->nr_cur_batched_writeback = nr_batch;
+	wb->nr_writeback_segs = nr_batch;
 
-	return r;
+	return err;
 }
 
 /*----------------------------------------------------------------------------*/
@@ -1166,7 +1179,7 @@ int try_alloc_writeback_ios(struct wb_device *wb, size_t nr_batch, gfp_t gfp)
 		wb->name = kthread_create( \
 				name##_proc, wb,  "dmwb_" #name); \
 		if (IS_ERR(wb->name)) { \
-			r = PTR_ERR(wb->name); \
+			err = PTR_ERR(wb->name); \
 			wb->name = NULL; \
 			DMERR("couldn't spawn " #name); \
 			goto bad_##name; \
@@ -1184,26 +1197,26 @@ int try_alloc_writeback_ios(struct wb_device *wb, size_t nr_batch, gfp_t gfp)
  */
 static int init_metadata(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 
-	r = init_segment_header_array(wb);
-	if (r) {
+	err = init_segment_header_array(wb);
+	if (err) {
 		DMERR("init_segment_header_array failed");
 		goto bad_alloc_segment_header_array;
 	}
 
-	r = ht_empty_init(wb);
-	if (r) {
+	err = ht_empty_init(wb);
+	if (err) {
 		DMERR("ht_empty_init failed");
 		goto bad_alloc_ht;
 	}
 
-	return r;
+	return err;
 
 bad_alloc_ht:
 	free_segment_header_array(wb);
 bad_alloc_segment_header_array:
-	return r;
+	return err;
 }
 
 static void free_metadata(struct wb_device *wb)
@@ -1214,13 +1227,13 @@ static void free_metadata(struct wb_device *wb)
 
 static int init_writeback_daemon(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 	size_t nr_batch;
 
 	atomic_set(&wb->writeback_fail_count, 0);
 	atomic_set(&wb->writeback_io_count, 0);
 
-	nr_batch = 8;
+	nr_batch = 32;
 	wb->nr_max_batched_writeback = nr_batch;
 	if (try_alloc_writeback_ios(wb, nr_batch, GFP_KERNEL))
 		return -ENOMEM;
@@ -1234,23 +1247,22 @@ static int init_writeback_daemon(struct wb_device *wb)
 	wb->force_drop = false;
 	CREATE_DAEMON(writeback_daemon);
 
-	return r;
+	return err;
 
 bad_writeback_daemon:
 	free_writeback_ios(wb);
-	return r;
+	return err;
 }
 
-static int init_flusher(struct wb_device *wb)
+static int init_flush_daemon(struct wb_device *wb)
 {
-	wb->flusher_wq = create_singlethread_workqueue("dmwb_flusher");
-	if (!wb->flusher_wq) {
-		DMERR("Failed to allocate flusher_wq");
-		return -ENOMEM;
-	}
-
+	int err = 0;
 	init_waitqueue_head(&wb->flush_wait_queue);
-	return 0;
+	CREATE_DAEMON(flush_daemon);
+	return err;
+
+bad_flush_daemon:
+	return err;
 }
 
 static int init_flush_barrier_work(struct wb_device *wb)
@@ -1267,96 +1279,96 @@ static int init_flush_barrier_work(struct wb_device *wb)
 
 static int init_writeback_modulator(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 	wb->writeback_threshold = 0;
 	CREATE_DAEMON(writeback_modulator);
-	return r;
+	return err;
 
 bad_writeback_modulator:
-	return r;
+	return err;
 }
 
 static int init_sb_record_updater(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 	wb->update_sb_record_interval = 0;
 	CREATE_DAEMON(sb_record_updater);
-	return r;
+	return err;
 
 bad_sb_record_updater:
-	return r;
+	return err;
 }
 
 static int init_data_synchronizer(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 	wb->sync_data_interval = 0;
 	CREATE_DAEMON(data_synchronizer);
-	return r;
+	return err;
 
 bad_data_synchronizer:
-	return r;
+	return err;
 }
 
 int resume_cache(struct wb_device *wb)
 {
-	int r = 0;
+	int err = 0;
 
 	wb->nr_segments = calc_nr_segments(wb->cache_dev, wb);
 	wb->nr_caches_inseg = (1 << (SEGMENT_SIZE_ORDER - 3)) - 1;
 	wb->nr_caches = wb->nr_segments * wb->nr_caches_inseg;
 
-	r = init_devices(wb);
-	if (r)
+	err = init_devices(wb);
+	if (err)
 		goto bad_devices;
 
-	r = init_metadata(wb);
-	if (r)
+	err = init_metadata(wb);
+	if (err)
 		goto bad_metadata;
 
-	r = init_writeback_daemon(wb);
-	if (r) {
+	err = init_writeback_daemon(wb);
+	if (err) {
 		DMERR("init_writeback_daemon failed");
 		goto bad_writeback_daemon;
 	}
 
-	r = recover_cache(wb);
-	if (r) {
+	err = recover_cache(wb);
+	if (err) {
 		DMERR("recover_cache failed");
 		goto bad_recover;
 	}
 
-	r = init_flusher(wb);
-	if (r) {
-		DMERR("init_flusher failed");
-		goto bad_flusher;
+	err = init_flush_daemon(wb);
+	if (err) {
+		DMERR("init_flush_daemon failed");
+		goto bad_flush_daemon;
 	}
 
-	r = init_flush_barrier_work(wb);
-	if (r) {
+	err = init_flush_barrier_work(wb);
+	if (err) {
 		DMERR("init_flush_barrier_work failed");
 		goto bad_flush_barrier_work;
 	}
 
-	r = init_writeback_modulator(wb);
-	if (r) {
+	err = init_writeback_modulator(wb);
+	if (err) {
 		DMERR("init_writeback_modulator failed");
 		goto bad_modulator;
 	}
 
-	r = init_sb_record_updater(wb);
-	if (r) {
+	err = init_sb_record_updater(wb);
+	if (err) {
 		DMERR("init_sb_recorder failed");
 		goto bad_updater;
 	}
 
-	r = init_data_synchronizer(wb);
-	if (r) {
+	err = init_data_synchronizer(wb);
+	if (err) {
 		DMERR("init_data_synchronizer failed");
 		goto bad_synchronizer;
 	}
 
-	return r;
+	return err;
 
 bad_synchronizer:
 	kthread_stop(wb->sb_record_updater);
@@ -1365,8 +1377,8 @@ bad_updater:
 bad_modulator:
 	destroy_workqueue(wb->barrier_wq);
 bad_flush_barrier_work:
-	destroy_workqueue(wb->flusher_wq);
-bad_flusher:
+	kthread_stop(wb->flush_daemon);
+bad_flush_daemon:
 bad_recover:
 	kthread_stop(wb->writeback_daemon);
 	free_writeback_ios(wb);
@@ -1375,7 +1387,7 @@ bad_writeback_daemon:
 bad_metadata:
 	free_devices(wb);
 bad_devices:
-	return r;
+	return err;
 }
 
 void free_cache(struct wb_device *wb)
@@ -1390,7 +1402,7 @@ void free_cache(struct wb_device *wb)
 
 	destroy_workqueue(wb->barrier_wq);
 
-	destroy_workqueue(wb->flusher_wq);
+	kthread_stop(wb->flush_daemon);
 
 	kthread_stop(wb->writeback_daemon);
 	free_writeback_ios(wb);
