@@ -104,7 +104,6 @@ struct segment_header_device {
 	 */
 	/* - FROM ------------------------------------ */
 	__le64 id;
-	/* TODO Add timestamp? */
 	__le32 checksum;
 	/*
 	 * The number of metablocks in this segment header to be considered in
@@ -149,23 +148,13 @@ struct segment_header {
 
 /*----------------------------------------------------------------------------*/
 
-/*
- * Foreground queues this object and flush daemon later pops one job to submit
- * logging write to the cache device.
- */
-struct flush_job {
-	struct work_struct work;
-	struct wb_device *wb;
-	struct segment_header *seg;
-	struct bio_list barrier_ios; /* List of deferred bios */
-};
-
 /*
  * RAM buffer is a buffer that any dirty data are first written into.
  */
 struct rambuffer {
+	struct segment_header *seg;
 	void *data;
-	struct flush_job job;
+	struct bio_list barrier_ios; /* List of deferred bios */
 };
 
 /*----------------------------------------------------------------------------*/
@@ -330,23 +319,23 @@ struct wb_device {
 
 	struct rambuffer *rambuf_pool;
 
+	atomic64_t last_queued_segment_id;
+
 	/*--------------------------------------------------------------------*/
 
 	/********************
 	 * One-shot Writeback
 	 ********************/
 
-	wait_queue_head_t writeback_mb_wait_queue;
 	struct dm_kcopyd_client *copier;
 
 	/*--------------------------------------------------------------------*/
 
-	/****************
-	 * Buffer Flusher
-	 ****************/
+	/**************
+	 * Flush Daemon
+	 **************/
 
-	mempool_t *flush_job_pool;
-	struct workqueue_struct *flusher_wq;
+	struct task_struct *flush_daemon;
 
 	/*
 	 * Wait for a specified segment to be flushed. Non-interruptible
@@ -401,14 +390,15 @@ struct wb_device {
 	atomic_t writeback_io_count;
 	atomic_t writeback_fail_count;
 
-	u32 nr_cur_batched_writeback;
 	u32 nr_max_batched_writeback; /* Tunable */
-	u32 nr_max_batched_writeback_copied;
+	u32 nr_max_batched_writeback_saved;
 
 	struct rb_root writeback_tree;
 
-	u32 num_writeback_segs; /* Number of segments to write back */
+	u32 nr_writeback_segs;
 	struct writeback_segment **writeback_segs;
+	u32 nr_cur_batched_writeback; /* Number of segments to be written back */
+	u32 nr_empty_segs;
 
 	/*--------------------------------------------------------------------*/
 
@@ -418,7 +408,7 @@ struct wb_device {
 
 	struct task_struct *writeback_modulator;
 	u8 writeback_threshold; /* Tunable */
-	u8 writeback_threshold_copied;
+	u8 writeback_threshold_saved;
 
 	/*--------------------------------------------------------------------*/
 
@@ -428,7 +418,7 @@ struct wb_device {
 
 	struct task_struct *sb_record_updater;
 	unsigned long update_sb_record_interval; /* Tunable */
-	unsigned long update_sb_record_interval_copied;
+	unsigned long update_sb_record_interval_saved;
 
 	/*--------------------------------------------------------------------*/
 
@@ -438,7 +428,7 @@ struct wb_device {
 
 	struct task_struct *data_synchronizer;
 	unsigned long sync_data_interval; /* Tunable */
-	unsigned long sync_data_interval_copied;
+	unsigned long sync_data_interval_saved;
 
 	/*--------------------------------------------------------------------*/
 
@@ -446,10 +436,12 @@ struct wb_device {
 	 * Read Caching
 	 **************/
 
+	u32 nr_read_cache_cells;
+	u32 nr_read_cache_cells_saved;
 	struct work_struct read_cache_work;
 	struct read_cache_cells *read_cache_cells;
 	u32 read_cache_threshold; /* Tunable */
-	u32 read_cache_threshold_copied;
+	u32 read_cache_threshold_saved;
 
 	/*--------------------------------------------------------------------*/
 
@@ -483,6 +475,8 @@ int prepare_overwrite(struct wb_device *, struct segment_header *, struct metabl
 
 /*----------------------------------------------------------------------------*/
 
+#define ASSERT(cond) BUG_ON(!(cond))
+
 #define check_buffer_alignment(buf) \
 	do_check_buffer_alignment(buf, #buf, __func__)
 void do_check_buffer_alignment(void *, const char *, const char *);
@@ -504,4 +498,22 @@ sector_t dm_devsize(struct dm_dev *);
 
 /*----------------------------------------------------------------------------*/
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,8,0)
+#define req_is_write(req) op_is_write((req)->bi_op)
+#define bio_is_barrier(bio) ((bio)->bi_opf & REQ_PREFLUSH)
+#define bio_is_fua(bio) ((bio)->bi_opf & REQ_FUA)
+#define WB_IO_WRITE .bi_op = REQ_OP_WRITE, .bi_op_flags = 0
+#define WB_IO_READ .bi_op = REQ_OP_READ, .bi_op_flags = 0
+#define WB_IO_WRITE_FUA .bi_op = REQ_OP_WRITE, .bi_op_flags = REQ_FUA
+#else
+#define req_is_write(req) ((req)->bi_rw == WRITE)
+#define bio_is_barrier(bio) ((bio)->bi_rw & REQ_FLUSH)
+#define bio_is_fua(bio) ((bio)->bi_rw & REQ_FUA)
+#define WB_IO_WRITE .bi_rw = WRITE
+#define WB_IO_READ .bi_rw = READ
+#define WB_IO_WRITE_FUA .bi_rw = WRITE_FUA
+#endif
+
+/*----------------------------------------------------------------------------*/
+
 #endif
